Require Import List.
Require Export Util Var Val CSet Map Exp Env AutoIndTac.

Set Implicit Arguments.

Open Scope map_scope.

(** * Intermediate Language IL/M *)
(* This file roughly corresponds to Chapter 3 (except for normalisation, which is defined in ILM_Equivalences.v) *)


(** [args] is the type of the list of variables passed at a goto ... *)
Definition args := list var.
(** ... while [params] is the type of the list of formal parameters *)
Definition params := list var.

Definition updE X `{Defaulted X} (Ecallee : env X) (Ecaller : env X) (Z : params) (Y:args)
  : env X := 
  update_with_list Z (lookup_list Ecaller Y) Ecallee.




(** Memory *)

Inductive cell: Type := (* Definition 2 *)
| C_Unalloc: cell
| C_Uninit: cell
| C: val -> cell.

Global Instance cell_eq_dec:EqDec cell eq.
Proof.
  intros x y. change ({x = y} + {x <> y}). decide equality. destruct_prop(v=v0); eauto.
Qed.

Definition mem := map val cell.
Definition mempty:mem := BMap.empty C_Unalloc.
Hint Unfold mem mempty.
Global Instance inst_defaulted_mem : Defaulted mem := Build_Defaulted mempty.

Definition merge_mem (m1 m2: mem) := mapjoin m1 m2 (domain m1). (* Definition 6 *)
Definition restrict_mem (m: mem) (c: cset val) := (* Definition 5*)
  update_list mempty (lookup m) (elements c).


Section stmt_annotation.
(* Parameterise the annotation, since the semantics does not depend on it *)
Variable Ann: Type.

(** ** Syntax *)
Inductive stmt : Type := (* Figure 3.1 *)
| stmtExp    (ann : Ann) (x : var) (e: exp) (b : stmt) : stmt
| stmtIf     (ann : Ann) (x : var) (b1 : stmt) (b2 : stmt) : stmt
| stmtGoto   (ann : Ann) (l : lab) (Y:args) (m: var) : stmt
| stmtReturn (ann : Ann) (x : var) : stmt
| stmtLet    (ann : Ann) (Z : params) (m : var) (s : stmt) (b : stmt) : stmt
| stmtAlloc  (ann : Ann) (a : var) (m : var) (b : stmt) : stmt
| stmtFree   (ann : Ann) (m' : var) (m : var) (a : var) (b : stmt) : stmt
| stmtStore  (ann : Ann) (m' : var) (m : var) (a : var) (x : var) (b : stmt) : stmt
| stmtLoad   (ann : Ann) (x : var) (m : var) (a : var) (b : stmt) : stmt
| stmtMerge  (ann : Ann) (m'' : var) (m : var) (m' : var) (b : stmt) : stmt
| stmtSplit  (ann : Ann) (m' : var) (m'' : var) (m : var) (A : cset var) (b : stmt).

Definition stmt_get_ann (s:stmt) := match s with
| stmtExp ann _ _ _ => ann
| stmtIf ann _ _ _ => ann
| stmtGoto ann _ _ _ => ann
| stmtReturn ann _ => ann
| stmtLet ann _ _ _ _ => ann
| stmtAlloc ann _ _ _ => ann
| stmtFree ann _ _ _ _ => ann
| stmtStore ann _ _ _ _ _ => ann
| stmtLoad ann _ _ _ _ => ann
| stmtMerge ann _ _ _ _ => ann
| stmtSplit ann _ _ _ _ _ => ann end.


(** ** Semantics *)
Inductive block : Type := (* Definition 3 (closure are called blocks here) *)
  blockI {
    block_E : env val;
    block_Z : params;
    block_m : var;
    block_s : stmt
  }.
Definition labenv := list block.

Definition state : Type := (env val * labenv * env mem * cset val * stmt)%type. (* Definition 4 *)
Hint Unfold labenv state.
(* value environment, block environment, store environment, used addresses, next statement *)
Definition state_E (s:state) := match s with (E, _, _, _, _) => E end.
Definition state_L (s:state) := match s with (_, L, _, _, _) => L end.
Definition state_S (s:state) := match s with (_, _, S', _, _) => S' end.
Definition state_U (s:state) := match s with (_, _, _, U, _) => U end.
Definition state_b (s:state) := match s with (_, _, _, _, b) => b end.


Definition empty_state b:state :=
  (empty, nil, empty, BSet.empty, b).


(** *** Functional Semantics *)

  (* Small-step semantics - Figure 3.3 *)
  Inductive fstep : state -> state -> Prop :=
  | fstepExp E L S U ann x e b v
    (def:exp_eval E e = Some v)
    : fstep (E, L, S, U, stmtExp ann x e b) (E[x<-v], L, S, U, b)

  | fstepIfT E L S U
    ann (x:var) (b1 b2 : stmt)
    (condTrue: val2bool (E[x]) = true)
    : fstep (E, L, S, U, stmtIf ann x b1 b2) (E, L, S, U, b1)
    
  | fstepIfF E L S U
    ann (x:var) (b1 b2:stmt)
    (condFalse:val2bool (E[x]) = false)
    : fstep (E, L, S, U, stmtIf ann x b1 b2) (E, L, S, U, b2)
    
  | fstepGoto E L S U ann l Y m
    blk (Ldef:get L (counted l) blk) E'
    (updOk:updE (block_E blk) E (block_Z blk) Y  = E')
    S' `(empty[(block_m blk) <- S[m]] = S')
    : fstep  (E,   L, S, U, stmtGoto ann l Y m)
            (E', drop (counted l) L, S', U, block_s blk)

  | fstepLet E L S U
    ann s Z m (b:stmt) 
    : fstep (E, L, S, U, stmtLet ann Z m s b) (E, (blockI E Z m s)::L, S, U, b)

  | fstepAlloc E L S U ann (a m:var) (b:stmt)
    (v:val) `(v = allocator U) (mv:mem) `(mv = mempty[v <- C_Uninit])
    : fstep (E, L, S, U, stmtAlloc ann a m b) (E[a <- v], L, S[m <- mv], U ∪ {{v}}, b)

  | fstepFree E L S U ann m' m a b
    `(S[m][E[a]] <> C_Unalloc) mv `(mv = S[m][E[a] <- C_Unalloc])
    : fstep (E, L, S, U, stmtFree ann m' m a b) (E, L, S[m' <- mv], U, b)

  | fstepStore E L S U ann m' m a x b
    `(S[m][E[a]] <> C_Unalloc) mv `(mv = S[m][E[a] <- C (E[x])])
    : fstep (E, L, S, U, stmtStore ann m' m a x b) (E, L, S[m' <- mv], U, b)

  | fstepLoad E L S U ann x m a b
    v `(S[m][E[a]] = C v)
    : fstep (E, L, S, U, stmtLoad ann x m a b) (E[x<-v], L, S, U, b)

  | fstepMerge E L S U ann (m'' m m':var) b
    `(domain (S[m]) ∩ domain (S[m']) = BSet.empty)
    mv `(mv = merge_mem (S[m]) (S[m']))
    : fstep (E, L, S, U, stmtMerge ann m'' m m' b) (E, L, S[m'' <- mv], U, b)

  | fstepSplit E L S U ann (m' m'' m:var) (A:cset var) (b:stmt)
     mv' `(mv' = restrict_mem (S[m]) (minus (domain (S[m])) (lookup_set E A)))
     mv'' `(mv'' = restrict_mem (S[m]) (lookup_set E A))
   : fstep (E, L, S, U, stmtSplit ann m' m'' m A b) (E, L, S[m' <- mv'][m'' <- mv''], U, b).

  (* Semantics are deterministic *)
  Lemma fstep_functional s s' s'' :
    fstep s s' -> fstep s s'' -> s' = s''.
  Proof.
    induction 1; inversion 1; intros; subst; repeat get_functional; try congruence.
  Qed.

  (* Big-step semantics - Figure 3.4 *)
  Inductive fevalsTo : state -> val -> Prop :=
  | fevalsToRet E L S U ann x v :
      E[x] = v  -> fevalsTo (E, L, S, U, stmtReturn ann x) v
  | fevalsToFstep s s' v :
      fstep s s' -> fevalsTo s' v -> fevalsTo s v.

  Lemma fevalsTo_functional s v v' :
    fevalsTo s v -> fevalsTo s v' -> v = v'.
  Proof.
    intros H. revert v'. induction H; intros v' H'; inv H'.
    - eauto with get.
    - inv H0.
    - inv H.
    - generalize (fstep_functional H1 H). intros. subst. auto.
  Qed.

  (* Reflexive, transitive closure of small-step relation *)
  Inductive fstepStar: state -> state -> Prop :=
  | fstepStarRefl st: fstepStar st st
  | fstepStarStep st1 st2 st3: fstep st1 st2 -> fstepStar st2 st3 -> fstepStar st1 st3.
  Lemma fstepStarTrans st1 st2 st3: fstepStar st1 st2 -> fstepStar st2 st3 -> fstepStar st1 st3.
  Proof.
  induction 1.
  - tauto.
  - intros steps1. specialize (IHfstepStar steps1). eapply fstepStarStep; eauto.
  Qed.
  Lemma fstep_fstepStar st1 st2: fstep st1 st2 -> fstepStar st1 st2.
  Proof.
  intros. eapply fstepStarStep.
  - eassumption.
  - apply fstepStarRefl.
  Qed.
  Lemma fevalsToFstepStar st st' v: fstepStar st st' -> fevalsTo st' v -> fevalsTo st v.
  Proof.
  intros step. induction step.
  - tauto.
  - intros evals. eapply fevalsToFstep; eauto.
  Qed.

  (* Characterisation of big-step semantics using ->* *)
  Lemma fevalsTo_star (st:state) v:
    fevalsTo st v <-> exists E L S U G x, fstepStar st (E, L, S, U, stmtReturn G x) /\ E[x] = v.
  Proof.
  split.
  - induction 1.
    + exists E. exists L. exists S. exists U. exists ann. exists x. split.
      * apply fstepStarRefl.
      * assumption.
    + destruct IHfevalsTo as [E [L [S [U [G [x [steps EQ]]]]]]]. exists E. exists L. exists S. exists U. exists G.
      exists x. split.
      * eapply fstepStarStep; eassumption.
      * assumption.
  - intros [E [L [S [U [G [x [steps EQ]]]]]]]. remember (stmtReturn G x) as b. remember (E, L, S, U, b) as st'. 
    induction steps.
    + rewrite Heqst', Heqb. apply fevalsToRet. assumption. 
    + eapply fevalsToFstep; eauto.
  Qed.

End stmt_annotation.

(* 
*** Local Variables: ***
*** coq-load-path: ("infra" "constr" "il" "isa") ***
*** End: ***
*)

