Require Import Le Plus List Coq.Arith.Compare Compare_dec Coq.Program.Tactics.
Require Import EnvTy ILM ILM_Types ILM_Preservation.

Open Scope map_scope.

Set Implicit Arguments.

(* This file contains definition and proofs of semantic equivalence:
   The main results (at the bottom) is of course Theorem 2 from the thesis. If a program is well-typed,
   then it is semantically equivalent to its normalised form.
   We also prove some trivial semantic equivalences like swapping stores. *)
(* This roughly corresponds to Section 5.2. *)

(** Equality on program states and type environments
    This is needed to deal with the extensionality issues *)
Definition eq_state (st1 st2:state):Prop := match st1, st2 with
| (E1, L1, S1, U1, b1), (E2, L2, S2, U2, b2) => E1 = E2 /\ L1 = L2 /\
                                                (forall m, default (S1[m]) = default (S2[m])) /\
                                                (forall m, S1[m] === S2[m]) /\
                                                U1 = U2 /\ b1 = b2
(* using map equality also for E is quite cumbersome when it coms to stmtExp:
   It would have to be extended to the label environments, which also must pairwise match. *)
end.

(* Equal states perform steps to equal states *)
Lemma fstep_sameState (st1 st1' st2:state): fstep st1 st1' -> eq_state st1 st2 ->
  exists st2', fstep st2 st2' /\ eq_state st1' st2'.
Proof.
destruct st1 as [[[[E1 L1] S1] U1] b1], st1' as [[[[E1' L1'] S1'] U1'] b1'].
destruct st2 as [[[[E2 L2] S2] U2] b2]. simpl.
intros step [EQE [EQL [EQSD [EQSV [EQU EQb]]]]]. subst.
assert (SAMEDOM:forall m, domain(S2[m]) = domain (S1[m])).
{ clear- EQSD EQSV. intros. cset_tac; rewrite domain_spec2 in *; crush. }
invc step.
- eexists. split; [eapply fstepExp|]; crush.
- eexists. split; [eapply fstepIfT|]; crush.
- eexists. split; [eapply fstepIfF|]; crush.
- eexists. split; [eapply fstepGoto|]; crush.
- eexists. split; [eapply fstepLet|]; crush.
- eexists. split; [eapply fstepAlloc|]; crush.
- eexists. split; [eapply fstepFree|]; crush.
- eexists. split; [eapply fstepStore|]; crush.
- eexists. split.
  + eapply fstepLoad. rewrite <-EQSV. eassumption.
  + crush.
- eexists. split.
  + eapply fstepMerge; crush.
  + repeat split.
    * intros. destruct_lookup_b. rewrite !merge_mem_default. eauto.
    * intros. destruct_lookup_b. apply merge_mem_equiv; eauto.
- eexists. split.
  + eapply fstepSplit; crush.
  + rewrite !SAMEDOM. repeat split.
    * intros. destruct_lookup; [|destruct_lookup_b]; rewrite !restrict_mem_default; reflexivity.
    * intros. destruct_lookup; [|destruct_lookup_b]; apply restrict_mem_equiv; eauto.
Qed.
(* The same for ->* *)
Lemma fsteps_sameState (st1 st1' st2:state): fstepStar st1 st1' -> eq_state st1 st2 ->
  exists st2', fstepStar st2 st2' /\ eq_state st1' st2'.
Proof.
intros steps. revert st2. induction steps; intros ? eqst.
- eexists. split.
  + eapply fstepStarRefl.
  + assumption.
- edestruct fstep_sameState with (st2:=st0) as [st0' [steps' eqst']].
  + eassumption.
  + exact eqst.
  + edestruct IHsteps with (st4:=st0') as [st2' [steps'' eqst'']].
    * assumption.
    * exists st2'. split; eauto using fstepStarTrans, fstep_fstepStar.
Qed.
(* And the same for the big-step semantics *)
Lemma fevalsTo_sameState (st1 st2:state) v: eq_state st1 st2 -> fevalsTo st1 v -> fevalsTo st2 v.
Proof.
intros eqst evals. apply fevalsTo_star in evals. apply fevalsTo_star.
destruct evals as [E' [L' [S' [U' [G [x [steps hasval]]]]]]]. edestruct fsteps_sameState as [st' [steps' eqst']].
- eexact steps.
- eexact eqst.
- destruct st' as [[[[E'' L''] S''] U''] b'']. destruct eqst' as [EQE [EQL [EQSD [EQSV [EQU EQb]]]]]. subst.
  exists E'', L'', S'', U'', G, x. split; [assumption|reflexivity].
Qed.

(* To argue about token correctness, it is okay to swap to an equal state during the execution *)
Lemma alias_tokens_correct_same_state te b rt: alias_tokens_correct te b rt ->
  forall s E L S U, stateConsistent te s (E, L, S, U, b) rt ->
  forall st' ost ost', fstepStar (E, L, S, U, b) st' -> eq_state ost st' -> fstepStar ost ost' ->
                       alias_tokens_apply (state_E ost') (state_G ost').
Proof.
intros correct ? ? ? ? ? types ? ? ? steps eqst osteps.
edestruct fsteps_sameState with (st2:=st') (st1:=ost) as [st'' [steps' eqst']]; try eassumption.
destruct ost' as [[[[Eo' Lo'] So'] Uo'] bo'], st'' as [[[[E'' L''] S''] U''] b''].
destruct eqst' as [EQE [EQL [EQSD [EQSV [EQU EQb]]]]]. subst Eo' Lo' Uo' bo'.
eapply correct with (st':=(E'', L'', S'', U'', b'')).
- eexact types.
- eapply fstepStarTrans; eassumption.
- reflexivity.
- reflexivity.
Qed.

(* Equalitiy on static contexts (type_env): Similar to the state equality above *)
Definition eq_type_env (te1 te2: type_env):Prop := match te1, te2 with
| (ET1, LT1, ST1, F1), (ET2, LT2, ST2, F2) => ET1 = ET2 /\ LT1 = LT2 /\ ST1 === ST2 /\ F1 = F2
(* again, map equality for ET makes things significantly more complicated *)
end.

(* A program well-typed under a static context is also well-typed under another context if both are
   equal *)
Lemma stmtOfType_sameEnv (te1 te2:type_env) (b:stmt) ty:
  stmtOfType te1 b ty -> eq_type_env te1 te2 -> stmtOfType te2 b ty.
Proof.
intros types. revert te2. induction types; destruct te2 as [[[ET2 LT2] ST2] F2]; simpl;
intros [EQET [EQLT [EQST EQF]]]; subst.
- eapply WT_Exp; crush.
- eapply WT_If; crush.
- eapply WT_Goto; crush.
- eapply WT_Return; crush.
- eapply WT_Let; crush.
- eapply WT_Alloc; eauto.
  + crush.
  + eapply IHtypes. crush.
- eapply WT_Free; eauto.
  + rewrite <- EQST. crush.
  + crush.
  + eapply IHtypes. crush.
- eapply WT_Store; eauto.
  + rewrite <- EQST. crush.
  + crush.
  + eapply IHtypes. crush.
- eapply WT_Load; eauto.
  + rewrite <- EQST. crush.
  + crush.
- eapply WT_MergeF; eauto.
  + rewrite <- EQST. crush.
  + rewrite <- EQST. crush.
  + rewrite <- EQST. crush.
  + eapply IHtypes. crush.
- eapply WT_MergeP; eauto.
  + rewrite <- EQST. eauto.
  + rewrite <- EQST. eauto.
  + rewrite <- EQST. eauto.
  + eapply IHtypes. crush.
- eapply WT_SplitF; eauto.
  + rewrite <- EQST. crush.
  + rewrite <- EQST. crush.
  + rewrite <- EQST. crush.
  + eapply IHtypes. crush.
- eapply WT_SplitP; eauto.
  + rewrite <- EQST. crush.
  + rewrite <- EQST. crush.
  + rewrite <- EQST. crush.
  + eapply IHtypes. crush.
Qed.


(** Definition of program equivalence: Definition 16 *)
Inductive equiv typeEnv rt (b1 b2: stmt):Prop :=
| equivC: (forall s E L S U v, stateConsistent typeEnv s (E, L, S, U, b1) rt ->
                               stateConsistent typeEnv s (E, L, S, U, b2) rt ->
                               (fevalsTo (E, L, S, U, b1) v <-> fevalsTo (E, L, S, U, b2) v)) ->
          (forall s E L S U, stateConsistent typeEnv s (E, L, S, U, b1) rt <->
                             stateConsistent typeEnv s (E, L, S, U, b2) rt) ->
           equiv typeEnv rt b1 b2.

(* Semantic equivalence is an equivalence relation *)
Lemma equiv_refl te rt b: equiv te rt b b.
Proof.
split; tauto.
Qed.
Lemma equiv_sym te rt b1 b2: equiv te rt b1 b2 -> equiv te rt b2 b1.
Proof.
intros [EQUIV EQSTATE]. split.
- intros. split; eapply EQUIV; eassumption.
- intros. split; eapply EQSTATE; eassumption.
Qed.
Lemma equiv_trans te rt b1 b2 b3: equiv te rt b1 b2 -> equiv te rt b2 b3 -> equiv te rt b1 b3.
Proof.
intros [EQUIV1 EQSTATE1] [EQUIV3 EQSTATE3]. split.
- intros. split; intros.
  + eapply EQUIV3, EQUIV1; try eassumption; eapply EQSTATE1; assumption.
  + eapply EQUIV1, EQUIV3; try eassumption; eapply EQSTATE3; assumption.
- intros. split; intros.
  + apply EQSTATE3, EQSTATE1. assumption.
  + apply EQSTATE1, EQSTATE3. assumption.
Qed.

(* Some lemmas useful to proof the second condition of semantic equivalence *)
(* It sufficces to show that both programs are well-typed under the same static contexts
    and have the same valid memory types *)
Lemma bothWelltypedValid_sameEnvs ET LT ST F rt (b1 b2: stmt):
  (stmtOfType (ET, LT, ST, F) b1 rt <-> stmtOfType (ET, LT, ST, F) b2 rt) ->
  (memTypesValid ET ST (stmt_get_ann b1) F <-> memTypesValid ET ST (stmt_get_ann b2) F) ->
  (forall s E L S U, stateConsistent (ET, LT, ST, F) s (E, L, S, U, b1) rt <->
                     stateConsistent (ET, LT, ST, F) s (E, L, S, U, b2) rt).
Proof.
intros stmt_types_eq mems_valid_eq ? ? ? ? ?. split; intros types.
- invc types. split; try assumption.
  + apply stmt_types_eq. assumption.
  + invc mems_type. split; try assumption. apply mems_valid_eq. assumption.
- invc types. split; try assumption.
  + apply stmt_types_eq. assumption.
  + invc mems_type. split; try assumption. apply mems_valid_eq. assumption.
Qed.

(* It sufficces to show that both programs are well-typed under the same static contexts
   and have the same alias annotation *)
Lemma bothWelltypedSameAlias_sameEnvs ET LT ST F rt (b1 b2: stmt):
  (stmtOfType (ET, LT, ST, F) b1 rt <-> stmtOfType (ET, LT, ST, F) b2 rt) ->
  stmt_get_ann b1 = stmt_get_ann b2 ->
  (forall s E L S U, stateConsistent (ET, LT, ST, F) s (E, L, S, U, b1) rt <->
                     stateConsistent (ET, LT, ST, F) s (E, L, S, U, b2) rt).
Proof.
intros stmt_types_eq alias_eq ? ? ? ? ?. apply bothWelltypedValid_sameEnvs.
- assumption.
- rewrite alias_eq. tauto.
Qed.


(** Swapping stores on independant memories preserves semantics *)
Lemma swapStore_sameSteps E L S U (b:stmt) G1 G2 (m1' m1 a1 x1 m2' m2 a2 x2:var):
  m1' <> m2 -> m2' <> m1 -> m1' <> m2' ->
  forall st1 st2, fstep (E, L, S, U, stmtStore G1 m1' m1 a1 x1 (stmtStore G2 m2' m2 a2 x2 b)) st1 ->
                  fstep st1 st2 ->
         exists st1' st2', fstep (E, L, S, U, stmtStore G2 m2' m2 a2 x2 (stmtStore G1 m1' m1 a1 x1 b)) st1' /\
                           fstep st1' st2' /\ eq_state st2 st2'.
Proof.
intros INEQ1 INEQ2 INEQ3 st1 st2 step1 step2.
invc step1. invc step2. eexists. eexists. split; [|split].
- eapply fstepStore; crush.
- eapply fstepStore; crush.
- repeat split; crush.
Qed.
Lemma swapStore_types te (b:stmt) rt G (m1' m1 a1 x1 m2' m2 a2 x2:var):
  m1' <> m2 -> m2' <> m1 -> m1' <> m2' -> m1 <> m2 ->
  stmtOfType te (stmtStore G m1' m1 a1 x1 (stmtStore G m2' m2 a2 x2 b)) rt ->
  stmtOfType te (stmtStore G m2' m2 a2 x2 (stmtStore G m1' m1 a1 x1 b)) rt.
Proof.
intros INEQ1 INEQ2 INEQ3 INEQ4 types. invc types. invc H13.
eapply WT_Store, WT_Store, stmtOfType_sameEnv; crush.
Qed.

Lemma swapStore_sameEvals E L S U (b:stmt) G m1' m1 a1 x1 m2' m2 a2 x2 v: m1' <> m2 -> m2' <> m1 -> m1' <> m2' ->
  fevalsTo (E, L, S, U, stmtStore G m1' m1 a1 x1 (stmtStore G m2' m2 a2 x2 b)) v ->
  fevalsTo (E, L, S, U, stmtStore G m2' m2 a2 x2 (stmtStore G m1' m1 a1 x1 b)) v.
Proof.
intros INEQ1 INEQ2 INEQ3 evals. inversion_clear evals. inv H. invc H0.
edestruct swapStore_sameSteps with (m1:=m1) (m1':=m1') (m2:=m2) (m2':=m2')
  as [st1' [st2' [step1' [step2' eqst]]]]; try eassumption.
eapply fevalsToFstep; [eassumption|]. eapply fevalsToFstep; [eassumption|].
eapply fevalsTo_sameState; eassumption.
Qed.

Lemma swapStore_alias_correct te (b:stmt) rt G m1' m1 a1 x1 m2' m2 a2 x2: 
  m1' <> m2 -> m2' <> m1 -> m1' <> m2' -> m1 <> m2 ->
  alias_tokens_correct te (stmtStore G m1' m1 a1 x1 (stmtStore G m2' m2 a2 x2 b)) rt ->
  alias_tokens_correct te (stmtStore G m2' m2 a2 x2 (stmtStore G m1' m1 a1 x1 b)) rt.
Proof.
intros INEQ1 INEQ2 INEQ3 INEQ4 correct. intros ? ? ? ? ? types ? ? ? steps EQE EQG. subst E' G'.
assert(types':stateConsistent te s (E, L, S, U, (stmtStore G m1' m1 a1 x1 (stmtStore G m2' m2 a2 x2 b))) rt).
{ invc types. split; try eassumption. eapply swapStore_types; eauto. }
invc steps.
- eapply correct; eauto using fstepStarRefl.
- inv H. invc H0.
  + eapply correct; eauto using fstepStarRefl.
  + inv H1. edestruct swapStore_sameSteps with (m1:=m2) (m1':=m2') (m2:=m1) (m2':=m1')
      as [st1' [st2' [step1' [step2' eqst]]]]; try (eassumption || (intros EQ; apply INEQ3; symmetry; assumption)).
    eapply (alias_tokens_correct_same_state correct) with (st':=st2').
    * eassumption.
    * eapply fstepStarStep, fstep_fstepStar; eassumption.
    * eexact eqst.
    * assumption.
Qed.

Theorem swapStore_equiv te (b:stmt) rt G m1' m1 a1 x1 m2' m2 a2 x2:
  m1' <> m2 -> m2' <> m1 -> m1' <> m2' -> m1 <> m2 ->
  equiv te rt (stmtStore G m1' m1 a1 x1 (stmtStore G m2' m2 a2 x2 b))
              (stmtStore G m2' m2 a2 x2 (stmtStore G m1' m1 a1 x1 b)).
Proof.
intros. split.
- intros. split; apply swapStore_sameEvals; eauto.
- destruct te as [[[ET LT] ST] F]. apply bothWelltypedSameAlias_sameEnvs.
  + split; eauto using swapStore_types.
  + simpl. reflexivity.
Qed.

(** Some more swaps: We only show that equal states are reached after two steps, and
    that well-typedness is preserved. The rest is obvious, but tedious to prove. *)
(* store and load *)
Lemma swapStoreLoad_sameSteps E L S U (b:stmt) G1 G2 G3 G4 (m1' m1 a1 x1 m2 a2 x2:var):
  m1' <> m2 -> x2 <> a1 -> x2 <> x1 ->
  forall st1 st2, fstep (E, L, S, U, stmtStore G1 m1' m1 a1 x1 (stmtLoad G2 x2 m2 a2 b)) st1 ->
                  fstep st1 st2 ->
         exists st1' st2', fstep (E, L, S, U, stmtLoad G3 x2 m2 a2 (stmtStore G4 m1' m1 a1 x1 b)) st1' /\
                           fstep st1' st2' /\ eq_state st2 st2'.
Proof.
intros INEQ1 INEQ2 INEQ3 st1 st2 step1 step2.
invc step1. invc step2. eexists. eexists. split; [|split].
- eapply fstepLoad; crush.
- eapply fstepStore; crush.
- repeat split; crush.
Qed.
Lemma swapStoreLoad_types te (b:stmt) rt G1 G2 (m1' m1 a1 x1 m2 a2 x2:var):
  m1' <> m2 -> m1 <> m2 ->
  stmtOfType te (stmtStore G1 m1' m1 a1 x1 (stmtLoad G2 x2 m2 a2 b)) rt ->
  stmtOfType te (stmtLoad G2 x2 m2 a2 (stmtStore G2 m1' m1 a1 x1 b)) rt.
Proof.
intros INEQ1 INEQ2 types. invc types. invc H13.
eapply WT_Load, WT_Store, stmtOfType_sameEnv; crush.
Qed.

(* load and store *)
Lemma swapLoadStore_sameSteps E L S U (b:stmt) G1 G2 G3 G4 (m1' m1 a1 x1 m2 a2 x2:var):
  m1' <> m2 -> x2 <> a1 -> x2 <> x1 ->
  forall st1 st2, fstep (E, L, S, U, stmtLoad G1 x2 m2 a2 (stmtStore G2 m1' m1 a1 x1 b)) st1 ->
                  fstep st1 st2 ->
         exists st1' st2', fstep (E, L, S, U, stmtStore G3 m1' m1 a1 x1 (stmtLoad G4 x2 m2 a2 b)) st1' /\
                           fstep st1' st2' /\ eq_state st2 st2'.
Proof.
intros INEQ1 INEQ2 INEQ3 st1 st2 step1 step2.
invc step1. invc step2. eexists. eexists. split; [|split].
- eapply fstepStore; crush.
- eapply fstepLoad; crush.
- repeat split; crush.
Qed.
Lemma swapLoadStore_types te (b:stmt) rt G (m1' m1 a1 x1 m2 a2 x2:var):
  m1' <> m2 -> m1 <> m2 -> x2 <> a1 -> x2 <> x1 ->
  stmtOfType te (stmtLoad G x2 m2 a2 (stmtStore G m1' m1 a1 x1 b)) rt ->
  stmtOfType te (stmtStore G m1' m1 a1 x1 (stmtLoad G x2 m2 a2 b)) rt.
Proof.
intros INEQ1 INEQ2 INEQ3 INEQ4 types. invc types. invc H11.
eapply WT_Store, WT_Load, stmtOfType_sameEnv; crush.
Qed.

(* free *)
Lemma swapFree_sameSteps E L S U (b:stmt) G1 G2 (m1' m1 a1 m2' m2 a2:var):
  m1' <> m2 -> m2' <> m1 -> m1' <> m2' ->
  forall st1 st2, fstep (E, L, S, U, stmtFree G1 m1' m1 a1 (stmtFree G2 m2' m2 a2 b)) st1 ->
                  fstep st1 st2 ->
         exists st1' st2', fstep (E, L, S, U, stmtFree G2 m2' m2 a2 (stmtFree G1 m1' m1 a1 b)) st1' /\
                           fstep st1' st2' /\ eq_state st2 st2'.
Proof.
intros INEQ1 INEQ2 INEQ3 st1 st2 step1 step2.
invc step1. invc step2. eexists. eexists. split; [|split].
- eapply fstepFree; crush.
- eapply fstepFree; crush.
- repeat split; crush.
Qed.
Lemma swapFree_types te (b:stmt) rt G (m1' m1 a1 m2' m2 a2:var):
  m1' <> m2 -> m2' <> m1 -> m1' <> m2' -> m1 <> m2 ->
  stmtOfType te (stmtFree G m1' m1 a1 (stmtFree G m2' m2 a2 b)) rt ->
  stmtOfType te (stmtFree G m2' m2 a2 (stmtFree G m1' m1 a1 b)) rt.
Proof.
intros INEQ1 INEQ2 INEQ3 INEQ4 types. invc types. invc H11.
eapply WT_Free, WT_Free, stmtOfType_sameEnv; crush.
Qed.

(* free and store *)
Lemma swapFreeStore_sameSteps E L S U (b:stmt) G1 G2 (m1' m1 a1 m2' m2 a2 x2:var):
  m1' <> m2 -> m2' <> m1 -> m1' <> m2' ->
  forall st1 st2, fstep (E, L, S, U, stmtFree G1 m1' m1 a1 (stmtStore G2 m2' m2 a2 x2 b)) st1 ->
                  fstep st1 st2 ->
         exists st1' st2', fstep (E, L, S, U, stmtStore G2 m2' m2 a2 x2 (stmtFree G1 m1' m1 a1 b)) st1' /\
                           fstep st1' st2' /\ eq_state st2 st2'.
Proof.
intros INEQ1 INEQ2 INEQ3 st1 st2 step1 step2.
invc step1. invc step2. eexists. eexists. split; [|split].
- eapply fstepStore; crush.
- eapply fstepFree; crush.
- repeat split; crush.
Qed.
Lemma swapFreeStore_types te (b:stmt) rt G (m1' m1 a1 m2' m2 a2 x2:var):
  m1' <> m2 -> m2' <> m1 -> m1' <> m2' -> m1 <> m2 ->
  stmtOfType te (stmtFree G m1' m1 a1 (stmtStore G m2' m2 a2 x2 b)) rt ->
  stmtOfType te (stmtStore G m2' m2 a2 x2 (stmtFree G m1' m1 a1 b)) rt.
Proof.
intros INEQ1 INEQ2 INEQ3 INEQ4 types. invc types. invc H11.
eapply WT_Store, WT_Free, stmtOfType_sameEnv; crush.
Qed.

(* free and load *)
Lemma swapFreeLoad_sameSteps E L S U (b:stmt) G1 G2 G3 G4 (m1' m1 a1 m2 a2 x2:var):
  m1' <> m2 -> x2 <> a1 ->
  forall st1 st2, fstep (E, L, S, U, stmtFree G1 m1' m1 a1 (stmtLoad G2 x2 m2 a2 b)) st1 ->
                  fstep st1 st2 ->
         exists st1' st2', fstep (E, L, S, U, stmtLoad G3 x2 m2 a2 (stmtFree G4 m1' m1 a1 b)) st1' /\
                           fstep st1' st2' /\ eq_state st2 st2'.
Proof.
intros INEQ1 INEQ2 st1 st2 step1 step2.
invc step1. invc step2. eexists. eexists. split; [|split].
- eapply fstepLoad; crush.
- eapply fstepFree; crush.
- repeat split; crush.
Qed.
Lemma swapFreeLoad_types te (b:stmt) rt G1 G2 (m1' m1 a1 m2 a2 x2:var):
  m1' <> m2 -> m1 <> m2 ->
  stmtOfType te (stmtFree G1 m1' m1 a1 (stmtLoad G2 x2 m2 a2 b)) rt ->
  stmtOfType te (stmtLoad G2 x2 m2 a2 (stmtFree G2 m1' m1 a1 b)) rt.
Proof.
intros INEQ1 INEQ2 types. invc types. invc H11.
eapply WT_Load, WT_Free, stmtOfType_sameEnv; crush.
Qed.


(** Normalisation of memory operations, and their preservation of semantics *)
(* "increment" a variable name (m-hat in Figure 3.5) *)
Definition var_s (m:var) := match m with
| mkvar n => mkvar (S n) end.
Lemma var_s_ineq m: m <> var_s m.
Proof.
unfold var_s. destruct m. injection. clear. induction n; crush.
Qed.


Fixpoint normalise (m:var) (b:stmt) := match b with (* Figure 3.5 *)
| stmtExp G x e b => stmtExp G x e (normalise m b)
| stmtIf G x b1 b2 => stmtIf G x (normalise m b1) (normalise m b2)
| stmtGoto G l Y _ => stmtGoto G l Y m
| stmtReturn G x => stmtReturn G x
| stmtLet G Z _ s b => stmtLet G Z m (normalise m s) (normalise m b)
| stmtAlloc G a _ b => stmtAlloc G a (var_s m) 
                      (stmtMerge (stmt_get_ann b) m (var_s m) m (normalise m b))
| stmtFree G _ _ a b => stmtFree G m m a (normalise m b)
| stmtStore G _ _ a x b => stmtStore G m m a x (normalise m b)
| stmtLoad G x _ a b => stmtLoad G x m a (normalise m b)
| stmtMerge G _ _ _ b => normalise m b
| stmtSplit G _ _ _ A b => normalise m b
end.

(* Normalisation preserves alias information: Lemma 4 *)
Lemma normalise_alias_same te m b rt: stmtOfType te b rt ->
  stmt_get_ann b = stmt_get_ann (normalise m b).
Proof.
revert te. induction b; intros te types; simpl in *; try tauto;
(* remaining cases: merge and split *) invc types; eapply IHb; crush.
Qed.

(* Normalisation preserves well-typedness: Lemma 5 *)
Lemma normalise_types ET LT ST F m (b:stmt) rt:
  stmtOfType (ET, LT, ST, F) b rt ->
  stmtOfType (ET, LT, ompty[m <- Some MemPanorama], BSet.empty) (normalise m b) rt.
Proof.
assert (INEQ:= @var_s_ineq m).
revert ET LT ST F rt. induction b; simpl; intros ? ? ? ? ? types.
- invc types. eapply WT_Exp; eauto.
  + erewrite <-normalise_alias_same; eassumption.
- invc types. eapply WT_If; eauto; erewrite <-normalise_alias_same; eassumption.
- invc types. eapply WT_Goto; eauto.
  + crush.
- invc types. eapply WT_Return; eauto.
- invc types. eapply WT_Let; eauto.
 + erewrite <-normalise_alias_same; eassumption.
- invc types. eapply WT_Alloc, WT_MergeP; simplify lookup'; eauto.
  + crush.
  + crush.
  + erewrite <-normalise_alias_same by eassumption. reflexivity.
  + eapply stmtOfType_sameEnv; [eapply IHb; crush|]. repeat split.
    * crush.
    * rewrite union_comm. apply union_minus. crush.
- invc types. eapply WT_Free; simplify lookup'; eauto.
  + crush.
  + erewrite <-normalise_alias_same by eassumption. assumption.
  + eapply stmtOfType_sameEnv; [eapply IHb; crush|]. repeat split.
    * crush.
- invc types. eapply WT_Store; simplify lookup'; eauto.
  + crush.
  + erewrite <-normalise_alias_same by eassumption. assumption.
  + eapply stmtOfType_sameEnv; [eapply IHb; crush|]. repeat split.
    * crush.
- invc types. eapply WT_Load; eauto.
  + crush.
  + crush.
  + erewrite <-normalise_alias_same by eassumption. assumption.
- invc types; eapply IHb; crush.
- invc types; eapply IHb; crush.
Qed.

(* Definition 18: When is a state the normalised form of another state *)
Definition normalise_block (m:var) (b:block _) := match b with
| blockI E Z _ b => blockI E Z m (normalise m b)
end.
Definition normalised_state (m:var) (te:type_env) (st:state) (nst:state):Prop := match te, st, nst with
| (_, LT, ST, _), (E1, L1, S1, U1, b1), (E2, L2, S2, U2, b2) =>
    (E1 = E2) /\
    (forall n tb, get LT n tb -> exists blk1 blk2, get L1 n blk1 /\ get L2 n blk2 /\ normalise_block m blk1 = blk2) /\
    (forall m, default (S2[m]) = C_Unalloc) /\ (merge_all_mems ST S1 === S2[m]) /\
    (U1 = U2) /\
    (normalise m b1 = b2)
end.

(* normalised states have the same E as the original one *)
Lemma normalised_same_E m te st nst:
  normalised_state m te st nst -> state_E st = state_E nst.
Proof.
destruct te as [[[ET LT] ST] F], st as [[[[E L] S] U] b], nst as [[[[En Ln] Sn] Un] bn].
intros [EQE _]. simpl. assumption.
Qed.
(* Normalised states have the same G as the original one *)
Lemma normalised_same_G m te te' st nst rt:  normalised_state m te st nst -> stmtOfType te' (state_b st) rt ->
  state_G st = state_G nst.
Proof.
destruct te as [[[ET LT] ST] F], st as [[[[E L] S] U] b], nst as [[[[En Ln] Sn] Un] bn].
intros [_ [_ [_ [_ [_ EQB]]]]] types. subst bn. simpl.
eapply normalise_alias_same; eassumption.
Qed.

(* The normalised state to a consistent state is also consistent *)
Lemma normalise_state_types te ET LT s m (st nst:state) rt (isET: ET = type_env_ET te) (isLT: LT = type_env_LT te):
  normalised_state m te st nst ->
  stateConsistent te s st rt -> memoriesDisjoint (type_env_ST te) (state_S st) ->
  stateConsistent (ET, LT, ompty[m <- Some MemPanorama], BSet.empty) s nst rt.
Proof.
intros normalised types. invc types. destruct nst as [[[[En Ln] Sn] Un] bn].
destruct normalised as [EQE [EQL [EQSd [EQS [EQU EQB]]]]]. subst Un En bn. simpl in *.
split; try assumption.
- eapply normalise_types. eassumption.
- intros ? ? ? ? GET. destruct (EQL _ _ GET) as [blk1 [blk2 [GET1 [GET2 NORM]]]].
  destruct (labenv_types _ _ _ _ GET) as [blk [GET' [blk_stmt_types blk_env_types]]].
  rewrite (get_functional GET1 GET') in *. exists blk2. split; [assumption|].
  destruct blk, blk2. simpl in *. invc NORM. split.
  + eapply normalise_types. eassumption.
  + assumption.
- erewrite <-normalise_alias_same by eassumption. inv mems_type. eapply memPanoramaOnlyTypes; try reflexivity.
  + intros x IN. rewrite domain_spec2, EQSd, <-EQS in IN.
    assert (EQ: merge_all_mems ST S [x] = merge_all_mems ST S [x]) by reflexivity.
    rewrite merge_all_mems_hasval in EQ by eauto. destruct EQ as [m' [mt' [HASTY HASVAL]]].
    eapply mems_in_U.
    * eassumption.
    * rewrite domain_spec2. rewrite HASVAL, mems_default. assumption.
  + exact EQSd.
- eapply oracleSameMem with (mv1 := (merge_all_mems ST S)).
  + rewrite EQS. symmetry. eapply merge_all_mems_only_mem. assumption.
  + eapply oracleChangeLabs; try eassumption. intros ? ? ? GETLn GETLT.
    destruct (EQL _ _ GETLT) as [blk1 [blk2' [GET1 [GET2 NORM]]]].
    rewrite <-(get_functional GETLn GET2) in *. rewrite <-NORM. exists blk1. split; [assumption|].
    destruct blk1. reflexivity.
Qed.

(* If we have just a single memory and no closures, the normalised state is obtained by simply
   normalising the program. *)
Lemma welltypedSingleMem_normalised ET E L (S: env mem) U b m:
  (forall m', default (S[m']) = C_Unalloc) ->
  normalised_state m (ET, nil, ompty[m <- Some MemPanorama], BSet.empty) (E, L, S, U, b)
                     (E, L, S, U, normalise m b).
Proof.
intros def. repeat split.
- intros. inversion H.
- eauto.
- eapply merge_all_mems_only_mem; eauto.
Qed.

(* If a variable is accessible in a memory, that memory agrees with the normalised memory
   in the location given by that variable - Lemma 6 *)
Lemma in_mem_same_val ET LT (ST: onv memory_type) F E En L Ln (S Sn: env mem) U Un G b bn m m' mt a
  (hasty: ST[m'] = Some mt):
  memoriesOfType ET ST G F E S U -> alias_tokens_apply E G ->
  normalised_state m (ET, LT, ST, F) (E, L, S, U, b) (En, Ln, Sn, Un, bn) ->
  in_mem G F mt a -> S[m'][E[a]] = Sn[m][En[a]].
Proof.
intros mems_type tokens_apply [EQE [EQL [EQSd [EQS [EQU EQB]]]]] in_mem. subst En Un bn.
rewrite <-EQS. inv mems_type. destruct_prop(S [m'] [E [a]] = C_Unalloc).
- rewrite e. rewrite merge_all_mems_hasnoval by crush. intros ? ? hasty'. destruct_prop(m'=m0).
  + subst. rewrite domain_spec, mems_default. assumption.
  + eapply in_mem_not_in_other with (m1:=m') (m2:=m0); eauto.
- rewrite merge_all_mems_hasval by crush. exists m', mt. split.
  + assumption.
  + reflexivity.
Qed.



(* Having a state and its normalised form, jumping to a function preserves that relation. *)
Lemma normalised_goto ET LT ST F E L Ln (S Sn: env mem) U G b bn (m m':var) n blk ET' F' E' En' S' Sn'
  (updSame: E' = En') (hasty: ST[m'] = Some MemPanorama)
  (isbl: get L n blk) (memOk:empty[(block_m blk) <- S[m']] = S') (memOkn:empty[m <- Sn[m]] = Sn')
  (mems_type: memoriesOfType ET ST G BSet.empty E S U) (tokens_apply: alias_tokens_apply E G)
  (normalised:normalised_state m (ET, LT, ST, F) (E, L, S, U, b) (E, Ln, Sn, U, bn)):
  normalised_state m (ET', drop n LT, ompty[block_m blk <- Some MemPanorama], F')
                     (E', drop n L, S', U, block_s blk)
                     (En', drop n Ln, Sn', U, block_s (normalise_block m blk)).
Proof.
destruct normalised as [_ [EQL [EQSd [EQS [_ EQB]]]]].
destruct blk. simpl in *. repeat split.
- assumption.
- intros ? ? GET. apply get_drop in GET. destruct (EQL _ _ GET) as [blk1' [blk2' fits]].
  exists blk1', blk2'. split; [now apply drop_get|]. split; [now apply drop_get|]. tauto.
- simplify lookup'. crush.
- subst S' Sn'. simplify lookup'. rewrite <-EQS. symmetry.
  eapply merge_all_mems_one_mem; eauto using focusEmpty_memsEmpty.
Qed.

(* The address returned by the allocator is disjoint from the merged-all memory
   (needed to justify that the normalised program can perform the "merge"
   inserted after each "alloc") *)
Lemma normalised_alloc_disjoint ET LT ST F E L U S G b En Ln Sn Un bn m:
  normalised_state m (ET, LT, ST, F) (E, L, S, U, b) (En, Ln, Sn, Un, bn) ->
  memoriesOfType ET ST G F E S U -> alias_tokens_apply E G ->
  (domain (mempty[allocator Un <- C_Uninit])) ∩ (domain (Sn[m])) = BSet.empty.
Proof.
intros [EQE [EQL [EQSd [EQS [EQU EQB]]]]] mems_type tokens_apply. apply disjoint_equiv_meet_empty.
intros y [IN1 IN2]. rewrite domain_spec2 in *.
unfold mempty in *. rewrite default_spec2, default_spec in IN1. unfold mem in *. destruct_lookup_b.
clear IN1. rewrite EQSd, <-EQS in IN2. apply IN2. symmetry. rewrite merge_all_mems_hasnoval by eauto.
intros. eapply allocInNoMem; eauto.
Qed.

(* Having a state and its normalised form, performing an alloc and then a merge
   of the new memory into the merged-all one preserves that relation. *)
Lemma normalised_alloc ET LT ST F E L Ln S Sn U G b bn (a m' m m_s:var) b' v ET' E' En'
  (isnew:v = allocator U) (updSame: E' = En') (misnew: ST[m'] = None)
  (mems_type: memoriesOfType ET ST G F E S U) (tokens_apply: alias_tokens_apply E G)
  (normalised: normalised_state m (ET, LT, ST, F) (E, L, S, U, b) (E, Ln, Sn, U, bn)):
  normalised_state m (ET', LT, ST [m' <- Some (MemFocus {{a}})], F ∪ {{a}})
                     (E', L, S [m' <- mempty [v <- C_Uninit]], U ∪ {{v}}, b')
                     (En', Ln, Sn [m_s <- mempty [allocator U <- C_Uninit]]
                                  [m <- merge_mem (mempty [v <- C_Uninit]) (Sn [m])],
                      U ∪ {{v}}, normalise m b').
Proof.
destruct normalised as [_ [EQL [EQSd [EQS [_ EQB]]]]]. repeat split; try assumption.
- intros. destruct_lookup; [|destruct_lookup_b].
  + rewrite merge_mem_default. eauto.
  + unfold mempty. rewrite default_spec2, default_spec. reflexivity.
- simplify lookup'. rewrite <-merge_all_mems_new_mem by (eauto using allocInNoMem). intros y.
  destruct_lookup.
  + assert (IN: allocator U ∈ domain (mempty[allocator U <- C_Uninit])). { rewrite domain_spec2. crush. }
    unfold merge_mem. eapply mapjoin_spec in IN. rewrite IN. crush.
  + assert (NOTIN: y ∉ domain (mempty[allocator U <- C_Uninit])). { rewrite domain_spec. crush. }
    rewrite EQS. eapply mapjoin_spec in NOTIN. unfold merge_mem. rewrite NOTIN. reflexivity.
Qed.

(* Having a state and its normalised form, that relation is preserved by updating the memory
   at the location given by a variable *)
Lemma normalised_update ET LT ST F (E:env val) L Ln S Sn U G b bn (a m m' m'':var) v mt mt' b'
  (isnew: ST[m' <- None][m''] = None) (hasty: ST[m'] = Some mt) (hasval: S[m'][E[a]] <> C_Unalloc)
  (mems_type: memoriesOfType ET ST G F E S U) (tokens_apply: alias_tokens_apply E G)
  (normalised: normalised_state m (ET, LT, ST, F) (E, L, S, U, b) (E, Ln, Sn, U, bn)):
  normalised_state m (ET, LT, ST [m' <- None] [m'' <- Some mt'], F)
     (E, L, S [m'' <- S [m'] [E [a] <- v]], U, b')
     (E, Ln, Sn [m <- Sn [m] [E [a] <- v]], U, normalise m b').
Proof.
destruct normalised as [_ [EQL [EQSd [EQS [_ EQB]]]]]. inv mems_type. repeat split.
- assumption.
- intros. destruct_lookup; crush.
- simplify lookup'. rewrite <-EQS. symmetry.
  rewrite merge_all_mems_change_type with (mt1:=mt) by (invc types_valid; rewrite default_spec2; assumption).
  eapply merge_all_mems_rename_update; eauto.
Qed.


(* If a state makes a step to X, a corresponding normalised state makes a series of steps
   to a normalised state of X - Lemma 7. *)
Lemma normalise_orig_step te s E G st nst rt m (isE: E = state_E st) (isG: G = state_G st)
  (types: stateConsistent te s st rt) (tokens_apply: alias_tokens_apply E G)
  (normalised: normalised_state m te st nst): forall st', fstep st st' ->
  exists te' s', stateConsistent te' s' st' rt /\
                 exists nst', fstepStar nst nst' /\ normalised_state m te' st' nst'.
Proof.
intros ? step. inv types. destruct nst as [[[[En Ln] Sn] Un] bn].
assert (normalised':=normalised).
destruct normalised' as [EQE [EQL [EQSd [EQS [EQU EQB]]]]]. subst En Un bn.
inv step.
- edestruct (expPreservation types); try eassumption.
  eexists. eexists. split; [eassumption|]. eexists. split.
  + eapply fstep_fstepStar, fstepExp; crush.
  + repeat split; intros; crush.
- edestruct (ifPreservation types); try (reflexivity || eassumption).
  eexists. eexists. split; [eassumption|]. eexists. split.
  + eapply fstep_fstepStar, fstepIfT; crush.
  + repeat split; assumption.
- edestruct (ifPreservation types); try (reflexivity || eassumption).
  eexists. eexists. split; [eassumption|]. eexists. split.
  + eapply fstep_fstepStar, fstepIfF; crush.
  + repeat split; assumption.
- edestruct (gotoPreservation types) as [ET' types']; try (reflexivity || eassumption).
  eexists. eexists. split; [eassumption|]. invc stmt_types.
  destruct (EQL _ _ get0) as [blk1 [blk2 [get1 [get2 norm]]]]. assert (EQblk := get_functional Ldef get1).
  subst blk. eexists. split. 
  + eapply fstep_fstepStar, fstepGoto; eauto.
  + subst blk2. eapply normalised_goto; destruct blk1; crush.
- edestruct (letPreservation types); try (reflexivity || eassumption).
  eexists. eexists. split; [eassumption|]. eexists. split.
  + eapply fstep_fstepStar, fstepLet.
  + repeat split; try assumption. intros ? ? GET. invc GET.
    * eexists. eexists. split; [now eapply getLB|]. split; [now eapply getLB|]. simpl. reflexivity.
    * destruct (EQL _ _ H4) as [blk1 [blk2 fits]]. exists blk1, blk2. split; [constructor; tauto|].
      split; [constructor; tauto|]. tauto.
- assert (m <> var_s m) by eapply var_s_ineq.
  edestruct (allocPreservation types); try (reflexivity || eassumption).
  eexists. eexists. split; [eassumption|]. eexists. split.
  + eapply fstepStarStep, fstep_fstepStar; [eapply fstepAlloc|eapply fstepMerge]; try (reflexivity || eassumption).
    simplify lookup'. eauto using normalised_alloc_disjoint.
  + simplify lookup'. invc stmt_types. eapply normalised_alloc; eauto.
- edestruct (freePreservation types); try (reflexivity || eassumption).
  eexists. eexists. split; [eassumption|]. invc stmt_types. eexists. split.
  + eapply fstep_fstepStar, fstepFree; eauto.
    * erewrite <-in_mem_same_val; eauto.
  + eapply normalised_update; eauto.
- edestruct (storePreservation types); try (reflexivity || eassumption).
  eexists. eexists. split; [eassumption|]. invc stmt_types. eexists. split.
  + eapply fstep_fstepStar, fstepStore; eauto.
    * erewrite <-in_mem_same_val; eauto.
  + eapply normalised_update; eauto.
- edestruct (loadPreservation types); try (reflexivity || eassumption).
  eexists. eexists. split; [eassumption|]. eexists. split.
  + eapply fstep_fstepStar, fstepLoad; eauto.
    * invc stmt_types. erewrite <-in_mem_same_val; eauto.
  + repeat split; assumption.
- edestruct (mergePreservation types) as [mt [A types']]; try (reflexivity || eassumption).
  eexists. eexists. split; [eassumption|]. eexists. split.
  + eapply fstepStarRefl.
  + repeat split; try assumption.
    * symmetry. rewrite <-EQS. invc stmt_types; eapply merge_all_mems_merge; crush.
- edestruct (splitPreservation types) as [mt [F' types']]; try (reflexivity || eassumption).
  eexists. eexists. split; [eassumption|]. eexists. split.
  + eapply fstepStarRefl.
  + repeat split; try assumption.
    * symmetry. rewrite <-EQS. invc stmt_types; eapply merge_all_mems_split; crush.
Qed.

(* The same for ->*: If the original programs performs a series of steps, the
   normalised program can perform a series of steps ending in a corresponding normalised state. *)
Lemma normalise_orig_steps te s st nst rt m
  (types: stateConsistent te s st rt) (tokens: alias_tokens_correct te (state_b st) rt)
  (normalised: normalised_state m te st nst): forall st', fstepStar st st' ->
  exists te' s', stateConsistent te' s' st' rt /\
                 exists nst', fstepStar nst nst' /\ normalised_state m te' st' nst'.
Proof.
remember st as stIN. remember te as teIN. remember s as sIN. assert (typesIN := types). intros ? steps.
rewrite HeqstIN in steps. rewrite HeqstIN, HeqteIN in normalised. rewrite HeqstIN, HeqsIN, HeqteIN in types.
assert (stepsIN: fstepStar stIN st). { rewrite HeqstIN. apply fstepStarRefl. }
destruct stIN as [[[[E L] S] U] b]. specialize (tokens _ _ _ _ _ typesIN). simpl in *.
clear- stepsIN steps types tokens normalised. revert te s nst types normalised.
induction steps; intros.
- eexists. eexists. split; [eexact types|]. eexists. split.
  + eapply fstepStarRefl.
  + eassumption.
- edestruct normalise_orig_step with (st:=st1) as [te' [ss' [types' [nst' [steps' normalised']]]]]; try reflexivity.
  + eassumption.
  + eapply tokens; (eassumption||reflexivity).
  + eexact normalised.
  + eexact H.
  + edestruct IHsteps with (nst:=nst') as [te'' [s'' [types'' [nst'' [nsteps normalised'']]]]].
    * eapply fstepStarTrans; eauto using fstep_fstepStar.
    * eexact types'.
    * eexact normalised'.
    * eexists. eexists. split; [eexact types''|]. exists nst''. split; eauto using fstepStarTrans, fstep_fstepStar.
Qed.

(* For the other direction of semantic equivalence, some lemmas only function properly when all
   splits and merges in the original program are skipped first *)
Definition is_not_split_merge (b:stmt):= match b with
| stmtMerge _ _ _ _ _ => false
| stmtSplit _ _ _ _ _ _ => false
| _ => true
end.

(* Having a state and its normalised form, we can progress in the original execution until we
   are at a statement which is neither split nor merge, and the target state still corresponds to the
   original normalised state. *)
Lemma skip_merge_split_steps te s st nst rt m
  (types: stateConsistent te s st rt)
  (tokens_apply: alias_tokens_apply (state_E st) (state_G st))
  (normalised: normalised_state m te st nst):
  exists st' te', stateConsistent te' s st' rt /\ fstepStar st st' /\ is_not_split_merge (state_b st') /\
                  normalised_state m te' st' nst.
Proof.
destruct st as [[[[E L] S] U] b]. revert te s E L S U nst types tokens_apply normalised.
induction b; intros; try (eexists; exists te; split; [eassumption|]; split; [now apply fstepStarRefl|];
split; [simpl; constructor|]; assumption);
(* now only merge and split are left *)
destruct te as [[[ET LT] ST] F], nst as [[[[En Ln] Sn] Un] bn], normalised as [EQE [EQL [EQSd [EQS [EQU EQB]]]]];
subst En Un bn.
- edestruct (mergePreservation types) as [mt [A types']]; try (reflexivity || eassumption). invc types.
  edestruct (IHb _ _ _ _ _ _ (E, Ln, Sn, U, normalise m b) types') as
      [st' [te' [types'' [steps' [notSplitMerge normalised']]]]].
  + invc stmt_types; simpl in *; assumption.
  + repeat split; try assumption. rewrite <- EQS. symmetry. invc stmt_types; eapply merge_all_mems_merge; crush.
  + exists st', te'. repeat split; try assumption. eapply fstepStarStep; [|eexact steps']. eapply fstepMerge.
    * invc stmt_types; eapply disjoint_equiv_meet_empty, welltypedMemoriesAreDisjoint; crush.
    * reflexivity.
- edestruct (splitPreservation types) as [mt [F' types']]; try (reflexivity || eassumption). invc types.
  edestruct (IHb _ _ _ _ _ _ (E, Ln, Sn, U, normalise m b) types') as
      [st' [te' [types'' [steps' [notSplitMerge normalised']]]]].
  + invc stmt_types; simpl in *; assumption.
  + repeat split; try assumption. rewrite <- EQS. symmetry. invc stmt_types; eapply merge_all_mems_split; crush.
  + exists st', te'. repeat split; try assumption. eapply fstepStarStep; [|eexact steps']. eapply fstepSplit.
    * reflexivity.
    * reflexivity.
Qed.

(* An optional step: Either one step or no step, preserving E and G. Definition 19. *)
Inductive fstepOpt: state -> state -> Prop :=
| fstepOptNo st: fstepOpt st st
| fstepOptYes st1 st2: fstep st1 st2 -> state_E st1 = state_E st2 -> state_G st1 = state_G st2 -> fstepOpt st1 st2.

(* If a normalised state for X makes a step, then there is a normalised state zero or one steps down the
   normalised execution that's normalised for the successor of X. *)
Lemma normalise_normal_not_split_merge_step te s st nst rt m
  (types: stateConsistent te s st rt)
  (notSplitMerge: is_not_split_merge (state_b st))
  (tokens_apply: alias_tokens_apply (state_E st) (state_G st))
  (normalised: normalised_state m te st nst): forall nst', fstep nst nst' ->
  exists nst'', fstepOpt nst' nst'' /\
         exists st' s' te', stateConsistent te' s' st' rt /\ fstep st st' /\ normalised_state m te' st' nst''.
Proof.
intros ? step. inv types. destruct nst as [[[[En Ln] Sn] Un] bn].
assert (normalised':=normalised).
destruct normalised' as [EQE [EQL [EQSd  [EQS [EQU EQB]]]]]. subst En Un bn.
inv step; destruct b; simpl in notSplitMerge; try (exfalso; exact notSplitMerge || discriminate);
clear notSplitMerge; invc H4.
- edestruct (expPreservation types); try (reflexivity || eassumption).
  eexists. split; [now apply fstepOptNo|]. eexists. eexists. eexists. split; [eassumption|]. split.
  + eapply fstepExp; crush.
  + repeat split; assumption.
- edestruct (ifPreservation types) as [typesT _]; try (reflexivity || eassumption).
  eexists. split; [now apply fstepOptNo|]. eexists. eexists. eexists. split; [eassumption|]. split.
  + eapply fstepIfT; crush.
  + repeat split; assumption.
- edestruct (ifPreservation types) as [_ typesF]; try (reflexivity || eassumption).
  eexists. split; [now apply fstepOptNo|]. eexists. eexists. eexists. split; [eassumption|]. split.
  + eapply fstepIfF; crush.
  + repeat split; assumption.
- invc stmt_types. destruct (EQL _ _ get0) as [blk1 [blk2 [get1 [get2 normblk]]]].
  assert (EQblk := get_functional Ldef get2). subst blk.
  edestruct (gotoPreservation types); try (reflexivity || eassumption).
  eexists. split; [now apply fstepOptNo|]. eexists. eexists. eexists. split; [eassumption|]. split.
  + eapply fstepGoto; crush.
  + subst blk2. eapply normalised_goto; destruct blk1; crush.
- edestruct (letPreservation types); try (reflexivity || eassumption).
  eexists. split; [now apply fstepOptNo|]. eexists. eexists. eexists. split; [eassumption|]. split.
  + eapply fstepLet.
  + repeat split; try assumption. intros ? ? GET. invc GET.
    * eexists. eexists. split; [now eapply getLB|]. split; [now eapply getLB|]. simpl. reflexivity.
    * destruct (EQL _ _ H4) as [blk1 [blk2 fits]]. exists blk1, blk2. split; [apply getLS; tauto|].
      split; [apply getLS; tauto|]. tauto.
- assert (m <> var_s m) by eapply var_s_ineq.
  edestruct (allocPreservation types); try (reflexivity || eassumption).
  eexists. split.
  + eapply fstepOptYes.
    * eapply fstepMerge; try (reflexivity || eassumption). simplify lookup'.
      eauto using normalised_alloc_disjoint.
    * reflexivity.
    * unfold state_G. simpl. invc stmt_types. eapply normalise_alias_same. eassumption.
  + eexists; eexists; eexists; split; [eassumption|]. split.
    * eapply fstepAlloc; reflexivity.
    * simplify lookup'. invc stmt_types. eapply normalised_alloc; eauto.
- invc stmt_types. assert (S[m1][E[a0]] <> C_Unalloc). { erewrite in_mem_same_val; eauto. }
  edestruct (freePreservation types); try (reflexivity || eassumption).
  eexists. split; [now apply fstepOptNo|]. eexists. eexists. eexists. split; [eassumption|]. split.
  + eapply fstepFree; eauto.
  + eapply normalised_update; eauto.
- invc stmt_types. assert (S [m1] [E [a0]] <> C_Unalloc). { erewrite in_mem_same_val; eauto. }
  edestruct (storePreservation types); try (reflexivity || eassumption).
  eexists. split; [now apply fstepOptNo|]. eexists. eexists. eexists. split; [eassumption|]. split.
  + eapply fstepStore; eauto.
  + eapply normalised_update; eauto.
- invc stmt_types. assert (S[m1][E[a0]] = C v). { erewrite in_mem_same_val; eauto. }
  edestruct (loadPreservation types); try (reflexivity || eassumption).
  eexists. split; [now apply fstepOptNo|]. eexists. eexists. eexists. split; [eassumption|]. split.
  + eapply fstepLoad; eassumption.
  + repeat split; assumption.
Qed.

(* A lemma very similar to the previous one, which does not require the first state to be neither split nor merge.
   This is Lemma 8. *)
Lemma normalise_normal_step te s st nst rt m
  (types: stateConsistent te s st rt)
  (tokens_apply: alias_tokens_apply (state_E st) (state_G st))
  (normalised: normalised_state m te st nst): forall nst', fstep nst nst' ->
  exists nst'', fstepOpt nst' nst'' /\
         exists st' s' te', stateConsistent te' s' st' rt /\ fstepStar st st' /\ normalised_state m te' st' nst''.
Proof.
intros ? nstep.
destruct (skip_merge_split_steps _ _ types tokens_apply normalised) as
        [st' [te' [types' [steps' [notSplitMerge normalised']]]]].
edestruct normalise_normal_not_split_merge_step with (st:=st') (nst := nst) as
      [nst'' [nstepOpt [st'' [s'' [te'' [types'' [steps'' normalised'']]]]]]].
- eassumption.
- assumption.
- assert (EQE: state_E st = state_E st').
  { rewrite (normalised_same_E _ _ _ _ normalised). rewrite (normalised_same_E _ _ _ _ normalised').
    reflexivity. }
  assert (EQG: state_G st = state_G st').
  { invc types. rewrite (@normalised_same_G _ _ (ET, LT, ST, F) _ _ rt normalised) by assumption.
    invc types'. rewrite (@normalised_same_G _ _ (ET0, LT0, ST0, F0) _ _ rt normalised') by assumption.
    reflexivity. }
  rewrite <-EQE, <-EQG. assumption.
- eassumption.
- eexact nstep.
- exists nst''. split; [assumption|]. exists st'', s'', te''. split; [assumption|].
  split; [|assumption]. eauto using fstepStarTrans, fstep_fstepStar.
Qed.


(* The same for ->*: If the normalised program makes a series of steps, then (after optionally performing
   another normalised step) there is a corresponding original state which can be reached by a series of
   original steps, and that state is not a split or merge *)
Lemma normalise_normal_steps te s st nst rt m
  (types: stateConsistent te s st rt) (tokens: alias_tokens_correct te (state_b st) rt)
  (normalised: normalised_state m te st nst): forall nst', fstepStar nst nst' ->
  exists nst'', fstepOpt nst' nst'' /\
         exists st' s' te', stateConsistent te' s' st' rt /\ fstepStar st st' /\ normalised_state m te' st' nst''.
Proof.
intros nst' nsteps. destruct st as [[[[E L] S] U] b]. specialize (tokens _ _ _ _ _ types). simpl in *.
assert (Hmain: forall nst nst', fstepStar nst nst' -> forall te s rt st, fstepStar (E, L, S, U, b) st ->
               stateConsistent te s st rt -> forall nst'', fstepOpt nst nst'' -> normalised_state m te st nst'' ->
               exists nst''', fstepOpt nst' nst''' /\
                      exists st' s' te', stateConsistent te' s' st' rt /\ fstepStar st st' /\ 
                             normalised_state m te' st' nst''').
{ clear- tokens. intros ? ? nsteps. induction nsteps.
  - rename st into nst. intros ? ? ? ? steps types nst' nstepOpt normalised.
    exists nst'. split; [assumption|]. eexists. eexists. eexists.
    split; [eassumption|]. split; [|assumption]. constructor.
  - rename st1 into nst1. rename st2 into nst2. rename st3 into nst3.
    intros ? ? ? ? steps types nst' nstepOpt normalised.
    destruct nstepOpt; [|assert(EQst:=fstep_functional H H0); subst; eapply IHnsteps; now eauto using fstepOptNo].
    rename st0 into nst1. edestruct normalise_normal_step with (st:=st) (nst := nst1) as
      [nst4 [nstepOpt [st' [s' [te' [types' [steps' normalised']]]]]]].
    + eassumption.
    + eapply tokens with (st':=st); eauto using fstepStarTrans.
    + eexact normalised.
    + eexact H.
    + edestruct IHnsteps with (st:=st') (nst'' := nst4) as
          [nst'' [nstepOpt'' [st'' [s'' [te'' [typey'' [steps'' normalised'']]]]]]].
      * eauto using fstepStarTrans, fstep_fstepStar.
      * eassumption.
      * assumption.
      * eexact normalised'.
      * exists nst''. split; [assumption|]. exists st''. eexists. eexists.
        split; [eassumption|]. split; [|assumption]. eauto using fstepStarTrans, fstep_fstepStar. }
eapply (Hmain _ _ nsteps) with (nst'':=nst).
- apply fstepStarRefl.
- eassumption.
- apply fstepOptNo.
- assumption.
Qed.

(* If the alias tokens of the normalised program are correct, then so are the tokens
   of the original program - Corollary 2 *)
Lemma normalise_orig_alias_correct ET b rt m:
  alias_tokens_correct (ET, nil, ompty[m <- Some MemPanorama], BSet.empty) (normalise m b) rt ->
  alias_tokens_correct (ET, nil, ompty[m <- Some MemPanorama], BSet.empty) b rt.
Proof.
intros tokens_correct ? ? ? ? ? types st' ? ? steps. revert E' G'.
assert (normalised: normalised_state m (ET, nil, ompty [m <- Some MemPanorama], BSet.empty)
        (E, L, S, U, b) (E, L, S, U, normalise m b)).
{ invc types. invc mems_type. eapply welltypedSingleMem_normalised. eauto. }
assert (typesN: stateConsistent (ET, nil, ompty [m <- Some MemPanorama], BSet.empty) s (E, L, S, U, normalise m b) rt).
{ eapply normalise_state_types; try eassumption.
  - reflexivity.
  - reflexivity.
  - simpl. invc types. invc mems_type. split.
    + rewrite default_spec2. crush.
    + intros. unfold ompty in *. destruct_lookup; discriminate.
    + assumption. }
specialize (tokens_correct _ _ _ _ _ typesN).
remember (E, L, S, U, b) as st. remember (E, L, S, U, normalise m b) as nst.
remember (ET, nil, ompty [m <- Some MemPanorama], BSet.empty) as te.
assert (nsteps:fstepStar (E, L, S, U, normalise m b) nst) by (subst; apply fstepStarRefl).
clear- normalised types tokens_correct steps nsteps. revert te s nst normalised types nsteps.
(* Unfortunately, we have to do induction again and cannot use normalise_orig_steps - that lemma
   requires the original program to be alias correct, which we are only about to prove *)
induction steps; intros ? ? ? ? ? ? ? ? EQE' EQG'.
- eapply tokens_correct.
  + eexact nsteps.
  + erewrite EQE', normalised_same_E; eauto.
  + invc types. erewrite normalised_same_G; eauto.
- edestruct normalise_orig_step with (st:=st1) as [te' [ss' [types' [nst' [steps' normalised']]]]];
    try reflexivity.
  + eassumption.
  + simpl. eapply tokens_correct.
    * eexact nsteps.
    * eapply normalised_same_E. eassumption.
    * invc types. simpl. erewrite normalised_same_G; try eassumption. reflexivity.
  + eexact normalised.
  + eassumption.
  + eapply IHsteps with (nst:=nst'); try eassumption. eapply fstepStarTrans; eassumption.
Qed.

(* If the normalised program is alias-correct, then so is the original one - Corollary 3 *)
Lemma normalise_normal_alias_correct ET b rt m:
  stmtOfType (ET, nil, ompty[m <- Some MemPanorama], BSet.empty) b rt ->
  alias_tokens_correct (ET, nil, ompty[m <- Some MemPanorama], BSet.empty) b rt ->
  alias_tokens_correct (ET, nil, ompty[m <- Some MemPanorama], BSet.empty) (normalise m b) rt.
Proof.
intros stmt_types tokens_correct ? ? ? ? ? ntypes nst ? ? nsteps. invc ntypes.
assert (types: stateConsistent (ET, nil, ompty[m <- Some MemPanorama], BSet.empty) s (E, L, S, U, b) rt).
{ split; try eassumption. erewrite normalise_alias_same by eassumption. eassumption. }
edestruct normalise_normal_steps with (st:=(E, L, S, U, b)) (nst:=(E, L, S, U, normalise m b)) as 
    [nst' [nstepOpt [st' [s' [te' [types' [steps normalised]]]]]]].
- eexact types.
- exact tokens_correct.
- invc mems_type. eapply welltypedSingleMem_normalised. assumption.
- eexact nsteps.
- intros EQE' EQG'. subst E' G'. invc types'.
  assert (EQE:state_E nst = state_E nst'). { destruct nstepOpt; subst; auto. }
  assert (EQG:state_G nst = state_G nst'). { destruct nstepOpt; subst; auto. }
  rewrite EQE, EQG. eapply tokens_correct with (st':=(E0, L0, S0, U0, b0)).
  + eexact types.
  + eassumption.
  + erewrite <-normalised_same_E by eassumption. reflexivity.
  + erewrite <-normalised_same_G by eassumption. reflexivity.
Qed.

(* Core Theorem of our type theory: Theorem 2.
   If a program is well-typed and alias-correct, then it is semantically equivalent to
   its normalisation. *)
Theorem normalise_equiv ET b m rt: let te := (ET, nil, ompty[m <- Some MemPanorama], BSet.empty) in
  stmtOfType te b rt -> alias_tokens_correct te b rt -> equiv te rt b (normalise m b).
Proof.
intros te stmtTypes tokens. unfold te in *.  clear te. split.
- clear stmtTypes. intros ? ? ? ? ? ? types1 types2. split.
  + intros evals. apply fevalsTo_star in evals. apply fevalsTo_star.
    destruct evals as [E' [L' [S' [U' [G [x [steps hasval]]]]]]].
    edestruct normalise_orig_steps with (st' := (E', L', S', U', stmtReturn G x)) as
        [te' [ss' [_ [nst' [step' normalised']]]]].
    * eexact types1.
    * exact tokens.
    * invc types1. invc mems_type. eapply welltypedSingleMem_normalised. assumption.
    * exact steps.
    * destruct nst' as [[[[En' Ln'] Sn'] Un'] bn']. exists En', Ln', Sn', Un', G, x.
      destruct te' as [[[ET' LT'] ST'] F'], normalised' as [EQE [EQL [EQSd [EQS [EQU EQB]]]]].
      subst En' Un' bn'. split; assumption.
  + intros evals. apply fevalsTo_star in evals. apply fevalsTo_star.
    destruct evals as [En' [Ln' [Sn' [Un' [G [x [nsteps hasval]]]]]]].
    edestruct normalise_normal_steps with (nst' := (En', Ln', Sn', Un', stmtReturn G x)) as
        [nst' [nstepOpt' [st' [s' [te' [types' [steps' normalised']]]]]]].
    * eexact types1.
    * exact tokens.
    * invc types1. invc mems_type. eapply welltypedSingleMem_normalised. assumption.
    * exact nsteps.
    * invc nstepOpt'; [|now invc H]. specialize (tokens _ _ _ _ _ types1).
      destruct (skip_merge_split_steps _ _ types' (tokens _ _ _ steps' eq_refl eq_refl) normalised') as
        [st'' [te'' [types'' [steps'' [notSplitMerge'' normalised'']]]]].
      destruct st'' as [[[[E'' L''] S''] U''] b''], te'' as [[[ET'' LT''] ST''] F''].
      destruct normalised'' as [EQE'' [EQL'' [EQSd'' [EQS'' [EQU'' EQB'']]]]]. subst En' Un'.
      exists E'', L'', S'', U'', G, x. split; [|reflexivity].
      apply (fstepStarTrans steps'). apply (fstepStarTrans steps'').
      case_eq b''; intros; subst b''; simpl in *; try (exfalso; exact notSplitMerge'' || discriminate).
      rewrite EQB''. constructor.
- apply bothWelltypedSameAlias_sameEnvs.
  + split; eauto using normalise_types.
  + simpl. erewrite normalise_alias_same by eassumption. reflexivity.
Qed.

(*Print Assumptions normalise_fequiv.*)

(** Realisability *)

Inductive realisable ET m b rt: Prop := (* Definition 17 *)
| realisableC: (forall s E L S U, stateConsistent (ET, nil, ompty[m <- Some MemPanorama], BSet.empty) s
                                                  (E, L, S, U, b) rt ->
                forall v, fevalsTo (E, L, S, U, b) v -> fevalsTo (E, L, S, U, normalise m b) v) ->
               (exists s E L S U, stateConsistent (ET, nil, ompty[m <- Some MemPanorama], BSet.empty) s
                                                  (E, L, S, U, b) rt) ->
               realisable ET m b rt.

(* Well-typed, alias-correct programs are realisable: Corollary 1. *)
Lemma welltyped_realisable ET m b rt: stmtOfType (ET, nil, ompty[m <- Some MemPanorama], BSet.empty) b rt ->
  alias_tokens_correct (ET, nil, ompty[m <- Some MemPanorama], BSet.empty) b rt -> default ET = None ->
  realisable ET m b rt.
Proof.
intros types corr def. constructor.
- destruct (normalise_equiv types corr) as [EQUIV _].
  intros ? ? ? ? ? CONS v EVALS. specialize (EQUIV s E L S U v). apply EQUIV.
  + assumption.
  + invc CONS. constructor; try assumption.
    * eapply normalise_types. eassumption.
    * erewrite <-normalise_alias_same by eassumption. assumption.
  + assumption.
- apply synth_state.
  + assumption.
  + eapply memPanoramaOnlyValid.
  + assumption.
Qed.


(* 
*** Local Variables: ***
*** coq-load-path: ("infra" "constr" "il" "isa" ".") ***
*** End: ***
*)
